"""Tests for the package as a whole.

This module contains tests to check if everything from older package
versions is implemented and if the results are still the same.

This module contains the following functions:
get_all_statistics - get all implemented statistics
get_all_samplings - get all implemented samplings
sample_data - get sample data frame
sample_metadata - get sample metadata

This module contains the following tests:
test_all_statistics_and_samplings_from_old_versions_implemented
test_results_match_reference_results
"""

import ast
from collections import namedtuple
from configparser import ConfigParser
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from tests.test_metrics.create_sample_data_and_reference_results import (
    create_sample_data
)
from toarstats.metrics.interface import calculate_statistics
from toarstats.metrics.stats_utils import STATS_LIST


def get_all_statistics():
    """Get all implemented statistics.

    :return: A set of all implemented statistics
    """
    return set(STATS_LIST)


def get_all_samplings():
    """Get all implemented samplings.

    :return: A set of all implemented samplings
    """
    samplings = set()
    for file in Path(Path(__file__).resolve().parents[2],
                     "toarstats/metrics").glob("*.py"):
        for node in ast.parse(file.read_text(), file).body:
            if (isinstance(node, ast.Assign)
                    and isinstance(node.value, ast.Dict)
                    and node.targets[0].id == "RSTAGS"):
                samplings.update([sampling.value
                                  for sampling in node.value.keys])
    if "seasonal" in samplings and "vegseason" not in samplings:
        samplings.add("vegseason")
    return samplings


@pytest.fixture
def sample_data():
    """Get sample data frame.

    :return: A data frame with sample data
    """
    sample_data_file = Path(
        Path(__file__).resolve().parent, "sample_data/sample_data.csv"
    )
    if not sample_data_file.is_file():
        create_sample_data(sample_data_file.parent)
    return pd.read_csv(
        sample_data_file,
        header=None, index_col=0, parse_dates=True
    )


@pytest.fixture
def sample_metadata():
    """Get sample metadata.

    :return: A ``namedtuple`` with the sample metadata information
    """
    parser = ConfigParser()
    with open(Path(Path(__file__).resolve().parent,
                   "sample_data/sample_metadata.cfg"),
              encoding="utf-8") as file:
        parser.read_file(file)
    Metadata = namedtuple("Metadata", ["station_lat", "station_lon",
                                       "station_climatic_zone"])
    return Metadata(parser.getfloat("METADATA", "station_lat"),
                    parser.getfloat("METADATA", "station_lon"),
                    parser.getint("METADATA", "station_climatic_zone"))


def test_all_statistics_and_samplings_from_old_versions_implemented():
    """Test if all old statistics and samplings are implemented."""
    old_statistics = set()
    old_samplings = set()
    for file in Path(Path(__file__).resolve().parent,
                     "reference_results").rglob("*.csv"):
        statistic, sampling = file.stem.split("-")
        old_statistics.add(statistic)
        old_samplings.add(sampling)
    assert (old_statistics <= get_all_statistics()
            and old_samplings <= get_all_samplings())


@pytest.mark.parametrize("sampling", sorted(get_all_samplings()))
@pytest.mark.parametrize("statistic", sorted(get_all_statistics()))
def test_results_match_reference_results(statistic, sampling, sample_data,
                                         sample_metadata):
    """Test if the results match old package versions."""
    if statistic == "drmdmax1h" and sampling == "monthly":
        with pytest.raises(ValueError, match="The drmdmax1h statistic cannot"
                                             " be evaluated with monthly"
                                             " sampling"):
            result = calculate_statistics(sampling, statistic, sample_data,
                                          sample_metadata)
    else:
        result = calculate_statistics(sampling, statistic, sample_data,
                                      sample_metadata)
        for file in Path(
            Path(__file__).resolve().parent, "reference_results"
        ).rglob(f"{statistic}-{sampling}.csv"):
            expected_result = pd.read_csv(file, header=None, index_col=0,
                                          parse_dates=True)
            pd.testing.assert_index_equal(result.index, expected_result.index,
                                          check_names=False)
            if statistic == "w90" and sampling == "seasonal":
                with pytest.raises(AssertionError):
                    np.testing.assert_allclose(result.values,
                                               expected_result.values,
                                               rtol=1e-13)
            else:
                np.testing.assert_allclose(result.values,
                                           expected_result.values, rtol=1e-13)
