import numpy as np
import pandas as pd
import pytest

from toarstats.metrics.constants import (
    ALLOWED_SAMPLING_VALUES,
    RSTAGS,
    SEASON_DICT
)
from toarstats.metrics.defaults import DEFAULT_MIN_DATA_CAPTURE
from toarstats.metrics.interface import calculate_statistics
from toarstats.metrics.stats_utils import create_reference_series

data = pd.read_csv(
    "tests/test_metrics/time_series.csv", header=None, names=[None, "values"],
    index_col=0, parse_dates=True, infer_datetime_format=True
)
ref_data = create_reference_series(data.index)
metadata = {"station_lat": 50.906389,
            "station_lon": 6.403889,
            "station_climatic_zone": "cool temperate moist"}


@pytest.mark.slow
def test_value_count():
    for sampling in ALLOWED_SAMPLING_VALUES:
        result = calculate_statistics(sampling, "count", data, metadata)
        offset = RSTAGS["seasonal" if sampling == "vegseason" else sampling]
        expected = pd.DataFrame(index=result.index)
        if sampling in {"seasonal", "vegseason", "summer", "xsummer"}:
            for name in result.columns.to_list():
                months = SEASON_DICT[name.split("-", 1)[1]]
                cur_data = data[data.index.month.isin(months)]
                if 12 in months and 1 in months:
                    cur_data.index += pd.Timedelta(182, "days")
                expected[name] = cur_data.resample(offset).count()
        else:
            expected["count"] = data.resample(offset).count()
        expected = expected.astype("float64")
        pd.testing.assert_frame_equal(result, expected)


@pytest.mark.slow
def test_data_capture():
    for sampling in ALLOWED_SAMPLING_VALUES:
        result = calculate_statistics(sampling, "data_capture", data, metadata)
        offset = RSTAGS["seasonal" if sampling == "vegseason" else sampling]
        expected = pd.DataFrame(index=result.index)
        if sampling in {"seasonal", "vegseason", "summer", "xsummer"}:
            for name in result.columns.to_list():
                months = SEASON_DICT[name.split("-", 1)[1]]
                cur_data = data[data.index.month.isin(months)]
                cur_ref = ref_data[ref_data.index.month.isin(months)]
                if 12 in months and 1 in months:
                    cur_data.index += pd.Timedelta(182, "days")
                    cur_ref.index += pd.Timedelta(182, "days")
                count = cur_data.resample(offset).count().squeeze("columns")
                expected[name] = count.divide(
                    cur_ref.resample(offset).count()
                ).reindex(count.index)
        else:
            count = data.resample(offset).count().squeeze("columns")
            expected["data_capture"] = count.divide(
                ref_data.resample(offset).count()
            ).reindex(count.index)
        pd.testing.assert_frame_equal(result, expected)


@pytest.mark.slow
def test_average_values():
    for sampling in ALLOWED_SAMPLING_VALUES:
        result = calculate_statistics(sampling, "mean", data, metadata)
        offset = RSTAGS["seasonal" if sampling == "vegseason" else sampling]
        expected = pd.DataFrame(index=result.index)
        if sampling in {"seasonal", "vegseason", "summer", "xsummer"}:
            for name in result.columns.to_list():
                months = SEASON_DICT[name.split("-", 1)[1]]
                cur_data = data[data.index.month.isin(months)]
                if 12 in months and 1 in months:
                    cur_data.index += pd.Timedelta(182, "days")
                expected[name] = cur_data.resample(offset).mean()
                count = cur_data.resample(offset).count().reindex(
                    expected.index
                ).squeeze("columns")
                expected[name][count < 10] = np.nan
        else:
            expected["mean"] = data.resample(offset).mean()
            count = data.resample(offset).count().reindex(
                expected.index
            ).squeeze("columns")
            expected["mean"][count < 10] = np.nan
        pd.testing.assert_frame_equal(result, expected)


@pytest.mark.slow
def test_standard_deviation():
    for sampling in ALLOWED_SAMPLING_VALUES:
        result = calculate_statistics(sampling, "stddev", data, metadata)
        offset = RSTAGS["seasonal" if sampling == "vegseason" else sampling]
        expected = pd.DataFrame(index=result.index)
        if sampling in {"seasonal", "vegseason", "summer", "xsummer"}:
            for name in result.columns.to_list():
                months = SEASON_DICT[name.split("-", 1)[1]]
                cur_data = data[data.index.month.isin(months)]
                if 12 in months and 1 in months:
                    cur_data.index += pd.Timedelta(182, "days")
                expected[name] = cur_data.resample(offset).std()
                count = cur_data.resample(offset).count().reindex(
                    expected.index
                ).squeeze("columns")
                expected[name][count < 10] = np.nan
        else:
            expected["stddev"] = data.resample(offset).std()
            count = data.resample(offset).count().reindex(
                expected.index
            ).squeeze("columns")
            expected["stddev"][count < 10] = np.nan
        pd.testing.assert_frame_equal(result, expected)


@pytest.mark.slow
def test_median():
    for sampling in ALLOWED_SAMPLING_VALUES:
        result = calculate_statistics(sampling, "median", data, metadata)
        offset = RSTAGS["seasonal" if sampling == "vegseason" else sampling]
        expected = pd.DataFrame(index=result.index)
        if sampling in {"seasonal", "vegseason", "summer", "xsummer"}:
            for name in result.columns.to_list():
                months = SEASON_DICT[name.split("-", 1)[1]]
                cur_data = data[data.index.month.isin(months)]
                cur_ref = ref_data[ref_data.index.month.isin(months)]
                if 12 in months and 1 in months:
                    cur_data.index += pd.Timedelta(182, "days")
                    cur_ref.index += pd.Timedelta(182, "days")
                data_rs = cur_data.resample(offset)
                data_count = data_rs.count().squeeze("columns")
                ref_count = cur_ref.resample(offset).count()
                frac = data_count / ref_count
                tmp_res = data_rs.median().reindex(frac.index)
                tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
                expected[name] = tmp_res
        else:
            data_rs = data.resample(offset)
            data_count = data_rs.count().squeeze("columns")
            ref_count = ref_data.resample(offset).count()
            frac = data_count / ref_count
            tmp_res = data_rs.median().reindex(frac.index)
            tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
            expected["median"] = tmp_res
        pd.testing.assert_frame_equal(result, expected)


@pytest.mark.slow
def test_maximum():
    for sampling in ALLOWED_SAMPLING_VALUES:
        result = calculate_statistics(sampling, "max", data, metadata)
        offset = RSTAGS["seasonal" if sampling == "vegseason" else sampling]
        expected = pd.DataFrame(index=result.index)
        if sampling in {"seasonal", "vegseason", "summer", "xsummer"}:
            for name in result.columns.to_list():
                months = SEASON_DICT[name.split("-", 1)[1]]
                cur_data = data[data.index.month.isin(months)]
                cur_ref = ref_data[ref_data.index.month.isin(months)]
                if 12 in months and 1 in months:
                    cur_data.index += pd.Timedelta(182, "days")
                    cur_ref.index += pd.Timedelta(182, "days")
                data_rs = cur_data.resample(offset)
                data_count = data_rs.count().squeeze("columns")
                ref_count = cur_ref.resample(offset).count()
                frac = data_count / ref_count
                tmp_res = data_rs.max().reindex(frac.index)
                tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
                expected[name] = tmp_res
        else:
            data_rs = data.resample(offset)
            data_count = data_rs.count().squeeze("columns")
            ref_count = ref_data.resample(offset).count()
            frac = data_count / ref_count
            tmp_res = data_rs.max().reindex(frac.index)
            tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
            expected["max"] = tmp_res
        pd.testing.assert_frame_equal(result, expected)


@pytest.mark.slow
def test_minimum():
    for sampling in ALLOWED_SAMPLING_VALUES:
        result = calculate_statistics(sampling, "min", data, metadata)
        offset = RSTAGS["seasonal" if sampling == "vegseason" else sampling]
        expected = pd.DataFrame(index=result.index)
        if sampling in {"seasonal", "vegseason", "summer", "xsummer"}:
            for name in result.columns.to_list():
                months = SEASON_DICT[name.split("-", 1)[1]]
                cur_data = data[data.index.month.isin(months)]
                cur_ref = ref_data[ref_data.index.month.isin(months)]
                if 12 in months and 1 in months:
                    cur_data.index += pd.Timedelta(182, "days")
                    cur_ref.index += pd.Timedelta(182, "days")
                data_rs = cur_data.resample(offset)
                data_count = data_rs.count().squeeze("columns")
                ref_count = cur_ref.resample(offset).count()
                frac = data_count / ref_count
                tmp_res = data_rs.min().reindex(frac.index)
                tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
                expected[name] = tmp_res
        else:
            data_rs = data.resample(offset)
            data_count = data_rs.count().squeeze("columns")
            ref_count = ref_data.resample(offset).count()
            frac = data_count / ref_count
            tmp_res = data_rs.min().reindex(frac.index)
            tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
            expected["min"] = tmp_res
        pd.testing.assert_frame_equal(result, expected)


@pytest.mark.slow
def test_perc05():
    for sampling in ALLOWED_SAMPLING_VALUES:
        result = calculate_statistics(sampling, "p05", data, metadata)
        offset = RSTAGS["seasonal" if sampling == "vegseason" else sampling]
        expected = pd.DataFrame(index=result.index)
        if sampling in {"seasonal", "vegseason", "summer", "xsummer"}:
            for name in result.columns.to_list():
                months = SEASON_DICT[name.split("-", 1)[1]]
                cur_data = data[data.index.month.isin(months)]
                cur_ref = ref_data[ref_data.index.month.isin(months)]
                if 12 in months and 1 in months:
                    cur_data.index += pd.Timedelta(182, "days")
                    cur_ref.index += pd.Timedelta(182, "days")
                data_rs = cur_data.resample(offset)
                data_count = data_rs.count().squeeze("columns")
                ref_count = cur_ref.resample(offset).count()
                frac = data_count / ref_count
                tmp_res = data_rs.quantile(0.05).reindex(frac.index)
                tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
                expected[name] = tmp_res
        else:
            data_rs = data.resample(offset)
            data_count = data_rs.count().squeeze("columns")
            ref_count = ref_data.resample(offset).count()
            frac = data_count / ref_count
            tmp_res = data_rs.quantile(0.05).reindex(frac.index)
            tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
            expected["p05"] = tmp_res
        pd.testing.assert_frame_equal(result, expected)


@pytest.mark.slow
def test_perc10():
    for sampling in ALLOWED_SAMPLING_VALUES:
        result = calculate_statistics(sampling, "p10", data, metadata)
        offset = RSTAGS["seasonal" if sampling == "vegseason" else sampling]
        expected = pd.DataFrame(index=result.index)
        if sampling in {"seasonal", "vegseason", "summer", "xsummer"}:
            for name in result.columns.to_list():
                months = SEASON_DICT[name.split("-", 1)[1]]
                cur_data = data[data.index.month.isin(months)]
                cur_ref = ref_data[ref_data.index.month.isin(months)]
                if 12 in months and 1 in months:
                    cur_data.index += pd.Timedelta(182, "days")
                    cur_ref.index += pd.Timedelta(182, "days")
                data_rs = cur_data.resample(offset)
                data_count = data_rs.count().squeeze("columns")
                ref_count = cur_ref.resample(offset).count()
                frac = data_count / ref_count
                tmp_res = data_rs.quantile(0.1).reindex(frac.index)
                tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
                expected[name] = tmp_res
        else:
            data_rs = data.resample(offset)
            data_count = data_rs.count().squeeze("columns")
            ref_count = ref_data.resample(offset).count()
            frac = data_count / ref_count
            tmp_res = data_rs.quantile(0.1).reindex(frac.index)
            tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
            expected["p10"] = tmp_res
        pd.testing.assert_frame_equal(result, expected)


@pytest.mark.slow
def test_perc25():
    for sampling in ALLOWED_SAMPLING_VALUES:
        result = calculate_statistics(sampling, "p25", data, metadata)
        offset = RSTAGS["seasonal" if sampling == "vegseason" else sampling]
        expected = pd.DataFrame(index=result.index)
        if sampling in {"seasonal", "vegseason", "summer", "xsummer"}:
            for name in result.columns.to_list():
                months = SEASON_DICT[name.split("-", 1)[1]]
                cur_data = data[data.index.month.isin(months)]
                cur_ref = ref_data[ref_data.index.month.isin(months)]
                if 12 in months and 1 in months:
                    cur_data.index += pd.Timedelta(182, "days")
                    cur_ref.index += pd.Timedelta(182, "days")
                data_rs = cur_data.resample(offset)
                data_count = data_rs.count().squeeze("columns")
                ref_count = cur_ref.resample(offset).count()
                frac = data_count / ref_count
                tmp_res = data_rs.quantile(0.25).reindex(frac.index)
                tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
                expected[name] = tmp_res
        else:
            data_rs = data.resample(offset)
            data_count = data_rs.count().squeeze("columns")
            ref_count = ref_data.resample(offset).count()
            frac = data_count / ref_count
            tmp_res = data_rs.quantile(0.25).reindex(frac.index)
            tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
            expected["p25"] = tmp_res
        pd.testing.assert_frame_equal(result, expected)


@pytest.mark.slow
def test_perc75():
    for sampling in ALLOWED_SAMPLING_VALUES:
        result = calculate_statistics(sampling, "p75", data, metadata)
        offset = RSTAGS["seasonal" if sampling == "vegseason" else sampling]
        expected = pd.DataFrame(index=result.index)
        if sampling in {"seasonal", "vegseason", "summer", "xsummer"}:
            for name in result.columns.to_list():
                months = SEASON_DICT[name.split("-", 1)[1]]
                cur_data = data[data.index.month.isin(months)]
                cur_ref = ref_data[ref_data.index.month.isin(months)]
                if 12 in months and 1 in months:
                    cur_data.index += pd.Timedelta(182, "days")
                    cur_ref.index += pd.Timedelta(182, "days")
                data_rs = cur_data.resample(offset)
                data_count = data_rs.count().squeeze("columns")
                ref_count = cur_ref.resample(offset).count()
                frac = data_count / ref_count
                tmp_res = data_rs.quantile(0.75).reindex(frac.index)
                tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
                expected[name] = tmp_res
        else:
            data_rs = data.resample(offset)
            data_count = data_rs.count().squeeze("columns")
            ref_count = ref_data.resample(offset).count()
            frac = data_count / ref_count
            tmp_res = data_rs.quantile(0.75).reindex(frac.index)
            tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
            expected["p75"] = tmp_res
        pd.testing.assert_frame_equal(result, expected)


@pytest.mark.slow
def test_perc90():
    for sampling in ALLOWED_SAMPLING_VALUES:
        result = calculate_statistics(sampling, "p90", data, metadata)
        offset = RSTAGS["seasonal" if sampling == "vegseason" else sampling]
        expected = pd.DataFrame(index=result.index)
        if sampling in {"seasonal", "vegseason", "summer", "xsummer"}:
            for name in result.columns.to_list():
                months = SEASON_DICT[name.split("-", 1)[1]]
                cur_data = data[data.index.month.isin(months)]
                cur_ref = ref_data[ref_data.index.month.isin(months)]
                if 12 in months and 1 in months:
                    cur_data.index += pd.Timedelta(182, "days")
                    cur_ref.index += pd.Timedelta(182, "days")
                data_rs = cur_data.resample(offset)
                data_count = data_rs.count().squeeze("columns")
                ref_count = cur_ref.resample(offset).count()
                frac = data_count / ref_count
                tmp_res = data_rs.quantile(0.9).reindex(frac.index)
                tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
                expected[name] = tmp_res
        else:
            data_rs = data.resample(offset)
            data_count = data_rs.count().squeeze("columns")
            ref_count = ref_data.resample(offset).count()
            frac = data_count / ref_count
            tmp_res = data_rs.quantile(0.9).reindex(frac.index)
            tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
            expected["p90"] = tmp_res
        pd.testing.assert_frame_equal(result, expected)


@pytest.mark.slow
def test_perc95():
    for sampling in ALLOWED_SAMPLING_VALUES:
        result = calculate_statistics(sampling, "p95", data, metadata)
        offset = RSTAGS["seasonal" if sampling == "vegseason" else sampling]
        expected = pd.DataFrame(index=result.index)
        if sampling in {"seasonal", "vegseason", "summer", "xsummer"}:
            for name in result.columns.to_list():
                months = SEASON_DICT[name.split("-", 1)[1]]
                cur_data = data[data.index.month.isin(months)]
                cur_ref = ref_data[ref_data.index.month.isin(months)]
                if 12 in months and 1 in months:
                    cur_data.index += pd.Timedelta(182, "days")
                    cur_ref.index += pd.Timedelta(182, "days")
                data_rs = cur_data.resample(offset)
                data_count = data_rs.count().squeeze("columns")
                ref_count = cur_ref.resample(offset).count()
                frac = data_count / ref_count
                tmp_res = data_rs.quantile(0.95).reindex(frac.index)
                tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
                expected[name] = tmp_res
        else:
            data_rs = data.resample(offset)
            data_count = data_rs.count().squeeze("columns")
            ref_count = ref_data.resample(offset).count()
            frac = data_count / ref_count
            tmp_res = data_rs.quantile(0.95).reindex(frac.index)
            tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
            expected["p95"] = tmp_res
        pd.testing.assert_frame_equal(result, expected)


@pytest.mark.slow
def test_perc98():
    for sampling in ALLOWED_SAMPLING_VALUES:
        result = calculate_statistics(sampling, "p98", data, metadata)
        offset = RSTAGS["seasonal" if sampling == "vegseason" else sampling]
        expected = pd.DataFrame(index=result.index)
        if sampling in {"seasonal", "vegseason", "summer", "xsummer"}:
            for name in result.columns.to_list():
                months = SEASON_DICT[name.split("-", 1)[1]]
                cur_data = data[data.index.month.isin(months)]
                cur_ref = ref_data[ref_data.index.month.isin(months)]
                if 12 in months and 1 in months:
                    cur_data.index += pd.Timedelta(182, "days")
                    cur_ref.index += pd.Timedelta(182, "days")
                data_rs = cur_data.resample(offset)
                data_count = data_rs.count().squeeze("columns")
                ref_count = cur_ref.resample(offset).count()
                frac = data_count / ref_count
                tmp_res = data_rs.quantile(0.98).reindex(frac.index)
                tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
                expected[name] = tmp_res
        else:
            data_rs = data.resample(offset)
            data_count = data_rs.count().squeeze("columns")
            ref_count = ref_data.resample(offset).count()
            frac = data_count / ref_count
            tmp_res = data_rs.quantile(0.98).reindex(frac.index)
            tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
            expected["p98"] = tmp_res
        pd.testing.assert_frame_equal(result, expected)


@pytest.mark.slow
def test_perc99():
    for sampling in ALLOWED_SAMPLING_VALUES:
        result = calculate_statistics(sampling, "p99", data, metadata)
        offset = RSTAGS["seasonal" if sampling == "vegseason" else sampling]
        expected = pd.DataFrame(index=result.index)
        if sampling in {"seasonal", "vegseason", "summer", "xsummer"}:
            for name in result.columns.to_list():
                months = SEASON_DICT[name.split("-", 1)[1]]
                cur_data = data[data.index.month.isin(months)]
                cur_ref = ref_data[ref_data.index.month.isin(months)]
                if 12 in months and 1 in months:
                    cur_data.index += pd.Timedelta(182, "days")
                    cur_ref.index += pd.Timedelta(182, "days")
                data_rs = cur_data.resample(offset)
                data_count = data_rs.count().squeeze("columns")
                ref_count = cur_ref.resample(offset).count()
                frac = data_count / ref_count
                tmp_res = data_rs.quantile(0.99).reindex(frac.index)
                tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
                expected[name] = tmp_res
        else:
            data_rs = data.resample(offset)
            data_count = data_rs.count().squeeze("columns")
            ref_count = ref_data.resample(offset).count()
            frac = data_count / ref_count
            tmp_res = data_rs.quantile(0.99).reindex(frac.index)
            tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
            expected["p99"] = tmp_res
        pd.testing.assert_frame_equal(result, expected)


@pytest.mark.slow
def test_percentiles1():
    for sampling in ALLOWED_SAMPLING_VALUES:
        result = calculate_statistics(sampling, "percentiles1", data, metadata)
        offset = RSTAGS["seasonal" if sampling == "vegseason" else sampling]
        expected = pd.DataFrame(index=result.index)
        if sampling in {"seasonal", "vegseason", "summer", "xsummer"}:
            for name in result.columns.to_list():
                months = SEASON_DICT[name.split("-", 1)[1]]
                cur_data = data[data.index.month.isin(months)]
                cur_ref = ref_data[ref_data.index.month.isin(months)]
                if 12 in months and 1 in months:
                    cur_data.index += pd.Timedelta(182, "days")
                    cur_ref.index += pd.Timedelta(182, "days")
                data_rs = cur_data.resample(offset)
                data_count = data_rs.count().squeeze("columns")
                ref_count = cur_ref.resample(offset).count()
                frac = data_count / ref_count
                tmp_res = data_rs.quantile(
                    int(name.split("-", 1)[0][1:]) / 100
                ).reindex(frac.index)
                tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
                expected[name] = tmp_res
        else:
            data_rs = data.resample(offset)
            data_count = data_rs.count().squeeze("columns")
            ref_count = ref_data.resample(offset).count()
            frac = data_count / ref_count
            for name in result.columns.to_list():
                tmp_res = data_rs.quantile(int(name[1:]) / 100).reindex(
                    frac.index
                )
                tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
                expected[name] = tmp_res
        pd.testing.assert_frame_equal(result, expected)


@pytest.mark.slow
def test_percentiles2():
    for sampling in ALLOWED_SAMPLING_VALUES:
        result = calculate_statistics(sampling, "percentiles2", data, metadata)
        offset = RSTAGS["seasonal" if sampling == "vegseason" else sampling]
        expected = pd.DataFrame(index=result.index)
        if sampling in {"seasonal", "vegseason", "summer", "xsummer"}:
            for name in result.columns.to_list():
                months = SEASON_DICT[name.split("-", 1)[1]]
                cur_data = data[data.index.month.isin(months)]
                cur_ref = ref_data[ref_data.index.month.isin(months)]
                if 12 in months and 1 in months:
                    cur_data.index += pd.Timedelta(182, "days")
                    cur_ref.index += pd.Timedelta(182, "days")
                data_rs = cur_data.resample(offset)
                data_count = data_rs.count().squeeze("columns")
                ref_count = cur_ref.resample(offset).count()
                frac = data_count / ref_count
                tmp_res = data_rs.quantile(
                    int(name.split("-", 1)[0][1:]) / 100
                ).reindex(frac.index)
                tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
                expected[name] = tmp_res
        else:
            data_rs = data.resample(offset)
            data_count = data_rs.count().squeeze("columns")
            ref_count = ref_data.resample(offset).count()
            frac = data_count / ref_count
            for name in result.columns.to_list():
                tmp_res = data_rs.quantile(int(name[1:]) / 100).reindex(
                    frac.index
                )
                tmp_res[frac < DEFAULT_MIN_DATA_CAPTURE] = np.nan
                expected[name] = tmp_res
        pd.testing.assert_frame_equal(result, expected)
