"""Functions to check the user inputs.

This module contains the following functions:
is_correct_type - check if the parameter has the correct type
check_sampling - check the sampling parameter
is_empty - check if the parameter is an empty collection or string
to_collection - convert the parameter to a collection of strings
check_statistics - check the statistics parameter
from_pandas - extract the index and values from a data frame or series
check_index - check the index
check_values - check the values
check_data - check the data parameter
is_in_range - check if the parameter is in the given range
check_station_latitude - check the station latitude
check_station_longitude - check the station longitude
check_station_climatic_zone - check the station climatic zone
check_metadata - check the metadata parameter
check_seasons - check the seasons parameter
check_crops - check the crops parameter
check_min_data_capture - check the min data capture parameter
check_required_parameters - check that all required parameters are given
check_input_parameters - check all input parameters
"""

from collections import namedtuple

import numpy as np
import pandas as pd

from toarstats.metrics.constants import (
    ALLOWED_CROPS_VALUES, ALLOWED_SAMPLING_VALUES, CLIMATIC_ZONE, SEASON_DICT,
    STATISTICS_LIST
)


def is_correct_type(parameter, allowed_types, allowed_collection_types=None):
    """Check if the parameter has the correct type.

    If the parameter is a collection (list, set, tuple) it will also be
    checked if the elements have the correct type.

    :param parameter: the parameter to check
    :param allowed_types: a single type or a list of types considered
                          suitable for the parameter
    :param allowed_collection_types: a single type or a list of types
                                     considered suitable for the
                                     elements of the parameter
    :return: True if the parameter type and the element types (if any)
             are correct else False
    """
    return_val = isinstance(parameter, allowed_types)
    if return_val and isinstance(parameter, (list, set, tuple)):
        return_val = all(isinstance(el, allowed_collection_types)
                         for el in parameter)
    return return_val


def check_sampling(sampling_in):
    """Check the sampling parameter.

    :param sampling_in: the given sampling parameter

    :raises TypeError: raised if the sampling parameter is missing or
                       not a string
    :raises ValueError: raised if the sampling parameter is not in the
                        allowed sampling values

    :return: A string with the requested sampling rate
    """
    if not is_correct_type(sampling_in, str):
        raise TypeError("The sampling parameter is required and must be a"
                        " string")
    sampling_out = sampling_in.strip()
    if sampling_out not in ALLOWED_SAMPLING_VALUES:
        raise ValueError("The sampling parameter must be one of"
                         f" {ALLOWED_SAMPLING_VALUES}")
    return sampling_out


def is_empty(parameter):
    """Check if the parameter is an empty collection or string.

    :param parameter: the parameter to check

    :return: True if the parameter is an empty collection or string else
             False
    """
    return not parameter or (isinstance(parameter, str)
                             and not parameter.strip())


def to_collection(parameter):
    """Convert the parameter to a collection of strings.

    :param parameter: the parameter to convert

    :return: A list, set or tuple of strings
    """
    return [parameter] if isinstance(parameter, str) else parameter


def check_statistics(statistics_in):
    """Check the statistics parameter.

    :param statistics_in: the given statistics parameter

    :raises TypeError: raised if the statistics parameter is missing or
                       not a string or list, set or tuple of strings
    :raises ValueError: raised if

                          - an empty list, set, tuple or string is given
                          - any given name is not recognized
                          - the ``diurnal_cycle`` statistic is not given
                            alone

    :return: A list of strings with the statistics to calculate
    """
    if not is_correct_type(statistics_in, (list, set, tuple, str), str):
        raise TypeError("The statistics parameter is required and must either"
                        " be a single string or a collection of strings in the"
                        " form of a list, set or tuple")
    if is_empty(statistics_in):
        raise ValueError("No statistics given")
    statistics_out = []
    for stat in to_collection(statistics_in):
        stat = stat.strip()
        if stat in statistics_out:
            continue
        if stat in STATISTICS_LIST:
            statistics_out.append(stat)
        else:
            raise ValueError(f"{stat} is not a recognized statistic, available"
                             f" are {STATISTICS_LIST}")
    if "diurnal_cycle" in statistics_out and len(statistics_out) > 1:
        raise ValueError("Diurnal cycle can only be evaluated alone")
    return statistics_out


def from_pandas(pandas_obj):
    """Extract the index and values from a data frame or series.

    :param pandas_obj: the given data frame or series

    :return: An index and a values array
    """
    index_out = pandas_obj.index
    if isinstance(pandas_obj, pd.Series):
        values_out = pandas_obj.to_numpy()
    elif "value" in pandas_obj.columns.to_list():
        values_out = pandas_obj["value"].to_numpy()
    elif "values" in pandas_obj.columns.to_list():
        values_out = pandas_obj["values"].to_numpy()
    elif pandas_obj.shape[1] > 0:
        values_out = pandas_obj.iloc[:, 0].to_numpy()
    else:
        values_out = np.array([])
    return index_out, values_out


def check_index(index):
    """Check the index.

    :param index: the given index

    :raises TypeError: raised if the index is not a datetime index
    :raises ValueError: raised if the index is empty, has duplicates or
                        null values
    """
    if not is_correct_type(index, pd.DatetimeIndex):
        raise TypeError("No datetime index given")
    if index.empty or index.has_duplicates or index.hasnans:
        raise ValueError("The datetime index may not be empty, have duplicates"
                         " or null values")


def check_values(values):
    """Check the values.

    :param values: the given values

    :raises TypeError: raised if the values are not in a one-dimensional
                       float or int array
    :raises ValueError: raised if the values array is empty or only
                        contains null values
    """
    if (not is_correct_type(values, np.ndarray) or values.ndim != 1
            or not (np.issubdtype(values.dtype, np.floating)
                    or np.issubdtype(values.dtype, np.integer))):
        raise TypeError("Values must be given in a one-dimensional float or"
                        " int array")
    if np.isnan(values).all():
        raise ValueError("No values given")


def check_data(data_in, datetimes_in, values_in):
    """Check the data parameter.

    :param data_in: the given data parameter
    :param datetimes_in: the given datetimes parameter
    :param values_in: the given values parameter

    :raises TypeError: raised if

                         - the data parameter is given and not a data
                           frame or series
                         - the index is not a datetime index
                         - the values are not in a one-dimensional float
                           or int array
    :raises ValueError: raised if

                          - the index is empty, has duplicates or null
                            values
                          - the values array is empty or only contains
                            null values
                          - the index and values have different lengths

    :return: A series with a datetime index and float or int values
    """
    if data_in is not None:
        if isinstance(data_in, (pd.DataFrame, pd.Series)):
            index_out, values_out = from_pandas(data_in)
        else:
            raise TypeError("The data parameter must be a data frame or"
                            " series")
    else:
        index_out, values_out = from_pandas(pd.Series(values_in,
                                                      index=datetimes_in))
    check_index(index_out)
    check_values(values_out)
    if index_out.size != values_out.size:
        raise ValueError("Datetime index and values must have the same length")
    if index_out.tz:
        index_out = index_out.tz_localize(None)
    return pd.Series(values_out, index_out)


def is_in_range(parameter, min_valid, max_valid):
    """Check if the parameter is in the given range.

    :param parameter: the parameter to check
    :param min_valid: the minimum value allowed
    :param max_valid: the maximum value allowed

    :return: True if the parameter is in the given range else False
    """
    return min_valid <= parameter <= max_valid


def check_station_latitude(station_latitude):
    """Check the station latitude.

    :param station_latitude: the given station latitude

    :raises TypeError: raised if the station latitude is not a float or
                       int value
    :raises ValueError: raised if the station latitude is not between
                        -90 and 90
    """
    if not is_correct_type(station_latitude, (float, int)):
        raise TypeError("The station latitude must be a float or int value")
    if not is_in_range(station_latitude, -90, 90):
        raise ValueError("The station latitude must be between -90 and 90")


def check_station_longitude(station_longitude):
    """Check the station longitude.

    :param station_longitude: the given station longitude

    :raises TypeError: raised if the station longitude is not a float or
                       int value
    :raises ValueError: raised if the station longitude is not between
                        -180 and 180
    """
    if not is_correct_type(station_longitude, (float, int)):
        raise TypeError("The station longitude must be a float or int value")
    if not is_in_range(station_longitude, -180, 180):
        raise ValueError("The station longitude must be between -180 and 180")


def check_station_climatic_zone(station_climatic_zone):
    """Check the station climatic zone.

    :param station_climatic_zone: the given station climatic zone

    :raises TypeError: raised if the station climatic zone is not an int
                       or string
    :raises ValueError: raised if the station climatic zone is not in
                        the ``CLIMATIC_ZONE`` dictionary

    :return: A string with the given station climatic zone
    """
    if isinstance(station_climatic_zone, int):
        if station_climatic_zone not in CLIMATIC_ZONE:
            raise ValueError("The station climatic zone must be containted in"
                             f" {CLIMATIC_ZONE}")
        return CLIMATIC_ZONE[station_climatic_zone]
    if isinstance(station_climatic_zone, str):
        station_climatic_zone = station_climatic_zone.strip()
        if station_climatic_zone not in CLIMATIC_ZONE.values():
            raise ValueError("The station climatic zone must be containted in"
                             f" {CLIMATIC_ZONE}")
        return station_climatic_zone
    raise TypeError("The station climatic zone must be an int or string"
                    f" contained in {CLIMATIC_ZONE}")


def check_metadata(metadata_in, station_lat_in, station_lon_in,
                   station_climatic_zone_in):
    """Check the metadata parameter.

    :param metadata_in: the given metadata parameter
    :param station_lat_in: the given station latitude parameter
    :param station_lon_in: the given station longitude parameter
    :param station_climatic_zone_in: the given station climatic zone
                                     parameter

    :raises KeyError: raised if any key in the metadata parameter is not
                      recognized
    :raises TypeError: raised if

                         - the metadata parameter is given and not a
                           dictionary or named tuple
                         - the station latitude is given and not a float
                           or int value
                         - the station longitude is given and not a
                           float or int value
                         - the station climatic zone is given and not an
                           int or string
    :raises ValueError: raised if

                          - the station latitude is given and not
                            between -90 and 90
                          - the station longitude is given and not
                            between -180 and 180
                          - the station climatic zone is given and not
                            in the ``CLIMATIC_ZONE`` dictionary

    :return: A named tuple with the given metadata information or None
             if no information was given
    """
    keys = ["station_lat", "station_lon", "station_climatic_zone"]
    if metadata_in is not None:
        if isinstance(metadata_in, dict):
            tmp_dict = metadata_in
        elif (isinstance(metadata_in, tuple)
              and hasattr(metadata_in, "_asdict")):
            tmp_dict = metadata_in._asdict()
        else:
            raise TypeError("The metadata parameter must be a dictionary or"
                            " named tuple")
    else:
        tmp_dict = dict(zip(keys, [station_lat_in, station_lon_in,
                                   station_climatic_zone_in]))
    if not set(tmp_dict) <= set(keys):
        raise KeyError(f"The metadata parameter can only have keys in {keys}")
    tmp_dict = {key: val for key, val in tmp_dict.items() if val is not None}
    if not tmp_dict:
        return None
    if "station_lat" in tmp_dict:
        check_station_latitude(tmp_dict["station_lat"])
    if "station_lon" in tmp_dict:
        check_station_longitude(tmp_dict["station_lon"])
    if "station_climatic_zone" in tmp_dict:
        tmp_dict["station_climatic_zone"] = check_station_climatic_zone(
            tmp_dict["station_climatic_zone"]
        )
    return namedtuple("Metadata", tmp_dict)(**tmp_dict)


def check_seasons(seasons_in):
    """Check the seasons parameter.

    :param seasons_in: the given seasons parameter

    :raises TypeError: raised if the seasons parameter is given and not
                       a string or list, set or tuple of strings
    :raises ValueError: raised if the seasons parameter is given and
                          - an empty list, set, tuple or string is given
                          - any given season is not recognized

    :return: A list of strings with the given seasons or None if no
             seasons were given
    """
    if seasons_in is None:
        return None
    if not is_correct_type(seasons_in, (list, set, tuple, str), str):
        raise TypeError("The seasons parameter must either be a single string"
                        " or a collection of strings in the form of a list,"
                        " set or tuple")
    if is_empty(seasons_in):
        raise ValueError("No seasons given")
    seasons_out = []
    for season in to_collection(seasons_in):
        season = season.strip()
        if season in seasons_out:
            continue
        if season in SEASON_DICT:
            seasons_out.append(season)
        else:
            raise ValueError(f"{season} is not a recognized season, available"
                             f" are {SEASON_DICT}")
    return seasons_out


def check_crops(crops_in):
    """Check the crops parameter.

    :param crops_in: the given crops parameter

    :raises TypeError: raised if the crops parameter is given and not a
                       string or list, set or tuple of strings
    :raises ValueError: raised if the crops parameter is given and
                          - an empty list, set, tuple or string is given
                          - any given crop is not recognized

    :return: A list of strings with the given crops or None if no crops
             were given
    """
    if crops_in is None:
        return None
    if not is_correct_type(crops_in, (list, set, tuple, str), str):
        raise TypeError("The crops parameter must either be a single string or"
                        " a collection of strings in the form of a list, set"
                        " or tuple")
    if is_empty(crops_in):
        raise ValueError("No crops given")
    crops_out = []
    for crop in to_collection(crops_in):
        crop = crop.strip()
        if crop in crops_out:
            continue
        if crop in ALLOWED_CROPS_VALUES:
            crops_out.append(crop)
        else:
            raise ValueError(f"{crop} is not a recognized crop, available are"
                             f" {ALLOWED_CROPS_VALUES}")
    return crops_out


def check_min_data_capture(min_data_capture_in):
    """Check the min data capture parameter.

    :param min_data_capture_in: the given min data capture parameter

    :raises TypeError: raised if the min data capture parameter is given
                       and not a float or int value
    :raises ValueError: raised if the min data capture parameter is
                        given and not between 0 and 1

    :return: A float or int value between 0 and 1 or None if no data
             capture was given
    """
    if min_data_capture_in is None:
        return None
    if not is_correct_type(min_data_capture_in, (float, int)):
        raise TypeError("The min data capture parameter must be a float or int"
                        " value")
    if not is_in_range(min_data_capture_in, 0, 1):
        raise ValueError("The min data capture parameter must be between 0 and"
                         " 1")
    return min_data_capture_in


def check_required_parameters(sampling, statistics, metadata, seasons):
    """Check that all required parameters are given.

    :param sampling: the given sampling parameter
    :param statistics: the given statistics parameter
    :param metadata: the given metadata parameter
    :param seasons: the given seasons parameter

    :raises TypeError: raised if any required metadata information is
                       missing

    :return: A named tuple containing the information on which
             parameters are required
    """
    required = namedtuple(
        "Required",
        "station_lat, station_lon, station_climatic_zone, seasons, crops"
    )(*5*[False])
    if any(stat in {"dark_aot40", "dark_avg", "daylight_aot40", "daylight_avg"}
           for stat in statistics):
        required = required._replace(station_lat=True)
        required = required._replace(station_lon=True)
    if sampling in {"summer", "xsummer"}:
        required = required._replace(station_lat=True)
    if sampling in {"seasonal", "vegseason"}:
        required = required._replace(seasons=True)
        if (seasons is None and (sampling == "vegseason"
                                 or (sampling == "seasonal"
                                     and any("aot40" in stat or "w126" in stat
                                             for stat in statistics)))):
            required = required._replace(station_lat=True)
            required = required._replace(station_climatic_zone=True)
            required = required._replace(crops=True)
    if required.station_lat and not hasattr(metadata, "station_lat"):
        raise TypeError("The station latitude is required and must be a float"
                        " or int value between -90 and 90")
    if required.station_lon and not hasattr(metadata, "station_lon"):
        raise TypeError("The station longitude is required and must be a float"
                        " or int value between -180 and 180")
    if required.station_climatic_zone and not hasattr(metadata,
                                                      "station_climatic_zone"):
        raise TypeError("The station climatic zone is required and must be an"
                        f" int or string contained in {CLIMATIC_ZONE}")
    return required


def check_input_parameters(
    sampling_in, statistics_in, data_in, metadata_in, seasons_in, crops_in,
    min_data_capture_in, datetimes_in, values_in, station_lat_in,
    station_lon_in, station_climatic_zone_in
):
    """Check all input parameters.

    :param sampling_in: the given sampling parameter
    :param statistics_in: the given statistics parameter
    :param data_in: the given data parameter
    :param metadata_in: the given metadata parameter
    :param seasons_in: the given seasons parameter
    :param crops_in: the given crops parameter
    :param min_data_capture_in: the given min data capture parameter
    :param datetimes_in: the given datetimes parameter
    :param values_in: the given values parameter
    :param station_lat_in: the given station latitude parameter
    :param station_lon_in: the given station longitude parameter
    :param station_climatic_zone_in: the given station climatic zone
                                     parameter

    :raises KeyError: raised if any key in the metadata parameter is not
                      recognized
    :raises TypeError: raised if any parameter has the wrong type or any
                       required parameter is missing
    :raises ValueError: raised if

                          - any parameter has the wrong value
                          - the ``drmdmax1h`` statistic is given
                            together with ``monthly`` sampling

    :return: A named tuple containing all the processed input parameters
    """
    sampling_out = check_sampling(sampling_in)
    statistics_out = check_statistics(statistics_in)
    if sampling_out == "monthly" and "drmdmax1h" in statistics_out:
        raise ValueError("The drmdmax1h statistic cannot be evaluated with"
                         " monthly sampling")
    data_out = check_data(data_in, datetimes_in, values_in)
    metadata_out = check_metadata(metadata_in, station_lat_in, station_lon_in,
                                  station_climatic_zone_in)
    seasons_out = check_seasons(seasons_in)
    crops_out = check_crops(crops_in)
    min_data_capture_out = check_min_data_capture(min_data_capture_in)
    required = check_required_parameters(sampling_out, statistics_out,
                                         metadata_out, seasons_out)
    return namedtuple(
        "Parameters",
        "sampling, statistics, data, metadata, seasons, crops,"
        " min_data_capture,"
        " required"
    )(sampling_out, statistics_out, data_out, metadata_out, seasons_out,
      crops_out, min_data_capture_out, required)
