"""Collection of dictionaries and helper functions for the package.

This module contains the following functions:
register - decorator to register statistics
aot40stat - evaluate the sum of values above the given threshold
calc_data_capture - calculate the data capture fraction
create_reference_series - create a reference series
dma8_processor - calculate the daily maximum 8-hour running mean
get_elevation_angle - calculate the solar elevation angle
get_growing_season - get the correct season name
get_seasons - get the correct seasons for each statistic
harmonize_time - harmonize the datetime index of all given series
kth_highest - calculate the kth highest value of the given series
perc - calculate percentiles
prepare_data - prepare the result list
resample - resample the given series
resample_with_date - resample the given series
somo_processor - calculate the sum of excess of dma8
stat_processor_1 - calculate statistics
stat_processor_2 - calculate statistics
threshold_processor - evaluate counts above the given threshold
w126stat - evaluate the sum of values

This module contains the following variable:
STATS_LIST - dictionary which contains mapping of names to functions
"""

import numpy as np
import pandas as pd

from toarstats.metrics.constants import RSTAGS, SEASON_DICT
from toarstats.metrics.defaults import DEFAULT_SEASONS
from toarstats.metrics.solar_position import calc_zenith

STATS_LIST = {}


def register(name):
    """Decorator to register statistics.

    :param name: name to register the statistic under

    :return: The inner function
    """
    def register_function(func):
        """Add the statistic to the dictionary.

        :param func: statistics function to register

        :return: The statistics function
        """
        STATS_LIST[name] = func
        return func
    return register_function


def aot40stat(ser, ref, sampling, how, mincount=None, minfrac=0.75,
              month_list=None, threshold=0.):
    """Evaluate the sum of values above the given threshold.

    This function should be used in the ``stat_processor_1`` as ``func``
    argument to evaluate ``AOT40`` statistics.

    In contrast to the standard ``resample`` function, ``aot40stat``
    calculates sums above a given threshold and it rescales results if
    less than 100% but more than ``min_data_capture`` data values are
    present.

    :param ser: a series containing the values
    :param ref: reference series
    :param sampling: aggregation type
    :param how: aggregation function
    :param mincount: minimum number of values that must be present to
                     accept a statistical result, most functions will
                     use ``minfrac`` instead
    :param minfrac: minimum fraction of data values that must be present
                    in order to accept the result of a statistical
                    function, if it is set to None ``mincount`` will be
                    used instead
    :param month_list: a list of months for ``seasonal`` statistics
    :param threshold: threshold value above which values shall be
                      accumulated

    :return: A series with the sums of values above the given threshold
    """
    ser_adj = ser - threshold
    ser_adj[ser_adj < 0.] = 0.
    ser_adj_resample = ser_adj.resample(sampling)
    ser_adj_sum = ser_adj_resample.sum()
    count = ser_adj_resample.count()
    fcov = count / ref.resample(sampling).count()
    ser_res = ser_adj_sum / fcov
    ser_res[fcov < minfrac] = np.nan
    return ser_res.reindex(count.index)


def calc_data_capture(ser, ref, sampling, how, mincount=0, minfrac=None,
                      month_list=None):
    """Calculate the data capture fraction.

    :param ser: a series containing the values
    :param ref: reference series
    :param sampling: aggregation type
    :param how: aggregation function
    :param mincount: minimum number of values that must be present to
                     accept a statistical result, most functions will
                     use ``minfrac`` instead
    :param minfrac: minimum fraction of data values that must be present
                    in order to accept the result of a statistical
                    function, if it is set to None ``mincount`` will be
                    used instead
    :param month_list: a list of months for ``seasonal`` statistics

    :return: A series with the data capture fraction
    """
    ser_tmp = ser.resample(sampling).count()
    fcov = ser_tmp / ref.resample(sampling).count()
    return fcov.reindex(ser_tmp.index)


def create_reference_series(index):
    """Create a reference series.

    :param index: the given index

    :return: A series with a date range spanning from the beginning of
             the earliest given year to the ending of the latest given
             year and filled with zeros
    """
    min_date = index.min()
    max_date = index.max()
    start_date = f"{min_date.year}-01-01 00:{min_date.minute}"
    end_date = f"{max_date.year}-12-31 23:{max_date.minute}"
    reference_index = pd.date_range(start=start_date, end=end_date, freq="H")
    return pd.Series(0, reference_index)


def dma8_processor(ser, ref, mtype, metadata, seasons, min_data_capture,
                   tshift=0, hours=None, dname="", strict=False,
                   average=False):
    """Calculate the daily maximum 8-hour running mean.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value
    :param tshift: number of hours to shift the timestamps
    :param hours: hour interval in which to calculate the running mean
    :param dname: name of the specific ``dma8``
    :param strict: a boolean denoting whether to use the strict rules
                   when to save a running mean
    :param average: a boolean denoting whether to average the values

    :return: A list of dictionaries containing the processed data
    """
    if hours is None:
        hours = [0, 23]
    label = f"dma8{dname.strip().lower()}"
    k = 26 if dname.strip().lower().startswith("eu") else 4
    if mtype != "daily" and average:
        label = f"avg{label}"
    hour = ser.index.hour
    xh = (hour >= hours[0]) & (hour <= hours[1])
    hour = ref.index.hour
    xhref = (hour >= hours[0]) & (hour <= hours[1])
    res = prepare_data(ser[xh], ref[xhref], mtype, seasons, label)
    dcap = 6
    dcap2 = min_data_capture * (hours[1]-hours[0]+1)
    for r in res:
        wrk = r["ser"].reindex(r["ref"].index)
        wrk = wrk[r["ser"].index.min():r["ser"].index.max()]
        tmp = wrk.rolling(window=8, min_periods=dcap).mean().shift(tshift)
        dres = resample(tmp, None, "D", "max",
                        mincount=int(dcap2) if strict else 1)
        if mtype == "daily":
            r["ser"] = dres
        else:
            dref = r["ref"].resample("D").mean().dropna()
            if average:
                res_ser = resample(dres, dref, RSTAGS[mtype], "mean",
                                   minfrac=min_data_capture,
                                   month_list=r["month_list"])
            else:
                res_ser = kth_highest(dres, dref, mtype, k, min_data_capture)
            r["ser"] = res_ser
    return res


def get_elevation_angle(latitude, longitude, time):
    """Calculate the solar elevation angle.

    Times must be given in ``local`` or ``solar`` time.

    :param latitude: the station's latitude
    :param longitude: the station's longitude
    :param time: a datetime index

    :return: The solar elevation angle
    """
    z, _ = calc_zenith(time, latitude, longitude)
    return 90. - z


def get_growing_season(croptype, climatic_zone, latitude):
    """Get the correct season name.

    :param croptype: the crop type
    :param climatic_zone: the climatic zone
    :param latitude: the station's latitude

    :raises KeyError: raised if a growing season is unknown

    :return: The growing season matching the crop type, climatic zone
             and latitude
    """
    myclimzone = climatic_zone.replace(" ", "_")
    res = f"{croptype}-{myclimzone}-{'NH' if latitude >= 0. else 'SH'}"
    if (croptype == "wheat" and "cool_temperate" in myclimzone
            and latitude < 0.):
        res = (f"{croptype}-{myclimzone}_"
               f"{'le' if np.abs(latitude) <= 30. else 'gt'}30-SH")
    if res not in SEASON_DICT:
        raise KeyError(f"{res} not included in SEASON_DICT")
    return res


def get_seasons(sampling, statistics, metadata, seasons, crops,
                seasons_required, crops_required):
    """Get the correct seasons for each statistic.

    :param sampling: the given sampling parameter
    :param statistics: the given statistics parameter
    :param metadata: the given metadata parameter
    :param seasons: the given seasons parameter
    :param crops: the given crops parameter
    :param seasons_required: boolean denoting whether the seasons
                             parameter is required
    :param crops_required: boolean denoting whether the crops parameter
                           is required

    :raises KeyError: raised if a growing season is unknown

    :return: A list containing the seasons for each statistic
    """
    if sampling in {"summer", "xsummer"}:
        return len(statistics)*[[
            f"{'N' if metadata.station_lat > 0 else 'S'}H-"
            f"{'X' if sampling == 'xsummer' else ''}Summer"
        ]]
    if not seasons_required:
        return len(statistics)*[None]
    if seasons is not None:
        return len(statistics)*[seasons]
    if not crops_required:
        return len(statistics)*[DEFAULT_SEASONS]
    crops_season = [get_growing_season(crop, metadata.station_climatic_zone,
                                       metadata.station_lat) for crop in crops]
    if sampling == "vegseason":
        return len(statistics)*[crops_season]
    return [crops_season if "aot40" in stat or "w126" in stat
            else DEFAULT_SEASONS for stat in statistics]


def harmonize_time(ser_list, mtype, rsfreq=None):
    """Harmonize the datetime index of all given series.

    :param ser_list: a list of series
    :param mtype: aggregation type
    :param rsfreq: resampling frequency for the datetime index

    :return: A list of series with harmonized datetime indices
    """
    if rsfreq is None:
        rsfreq = RSTAGS[mtype]
    tmin = min(record["ser"].index.min() for record in ser_list)
    tmax = max(record["ser"].index.max() for record in ser_list)
    try:
        newindex = pd.date_range(tmin, tmax, freq=rsfreq)
    except ValueError:
        newindex = pd.DatetimeIndex([])
    for record in ser_list:
        record["ser"] = record["ser"].reindex(newindex)
    return ser_list


def kth_highest(ser, ref, mtype, k, min_data_capture):
    """Calculate the kth highest value of the given series.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param k: value to return
    :param min_data_capture: a fractional coverage value

    :return: A series with the kth highest value
    """
    def hk(x):
        """Calculate the kth highest value.

        :param x: a series

        :return: The kth highest value
        """
        res = x.sort_values().tail(k)
        res = res[-k] if res.size > (k-1) else np.nan
        return res

    res = ser.dropna().resample(RSTAGS[mtype]).apply(hk)
    ser_resample = ser.resample(RSTAGS[mtype])
    if not res.size:
        res = pd.Series(np.nan, ser_resample.asfreq().index)
    else:
        cov = ser_resample.count() / ref.resample(RSTAGS[mtype]).count()
        res[cov < min_data_capture] = np.nan
    return res


def perc(ser, ref, mtype, seasons, min_data_capture, p):
    """Calculate percentiles.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param seasons: season names
    :param min_data_capture: a fractional coverage value
    :param p: requested percentiles

    :return: A list of dictionaries containing the processed data
    """
    return stat_processor_1(f"p{p:02d}", ser, ref, mtype, seasons,
                            how=f"quantile{0.01*p}", minfrac=min_data_capture)


def prepare_data(ser, ref, mtype, seasons, label):
    """Prepare the result list.

    It contains a dictionary for each ``seasonal`` result. For
    ``daily``, ``monthly``, ``summer``, and ``annual`` statistics, there
    will be only one list element (except for ``mkey="diurnal_cycle"``).

    Data are shifted by 6 months if the season includes ``December`` and
    ``January`` so that data from one season will be found in the
    statistics of the more recent year.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param seasons: season names
    :param label: key under which the results will be saved

    :return: A list of dictionaries containing the processed data
    """
    res = []
    if mtype in {"seasonal", "summer", "xsummer"}:
        for s in seasons:
            mlist = SEASON_DICT[s]
            tmp = ser[ser.index.month.isin(mlist)]
            tmpref = ref[ref.index.month.isin(mlist)]
            if tmp.empty:
                tmp = pd.Series(np.nan, [tmpref.index[0]])
            elif 12 in mlist and 1 in mlist:
                tmp.index += pd.Timedelta(182, "days")
                tmpref.index += pd.Timedelta(182, "days")
            res.append({"name": f"{label}-{s}", "ser": tmp, "ref": tmpref,
                        "month_list": mlist})
    else:
        res.append({"name": label, "ser": ser.copy(), "ref": ref.copy(),
                    "month_list": None})
    return res


def resample(ser, ref, sampling, how, mincount=0, minfrac=None,
             month_list=None):
    """Resample the given series.

    :param ser: a series containing the values
    :param ref: reference series
    :param sampling: aggregation type
    :param how: aggregation function
    :param mincount: minimum number of values that must be present to
                     accept a statistical result, most functions will
                     use ``minfrac`` instead
    :param minfrac: minimum fraction of data values that must be present
                    in order to accept the result of a statistical
                    function, if it is set to None ``mincount`` will be
                    used instead
    :param month_list: a list of months for ``seasonal`` statistics

    :return: A series with the resampled data
    """
    ser_resample = ser.resample(sampling)
    count = ser_resample.count()
    if how == "count":
        ser_tmp = count.astype(float)
    elif how.startswith("quantile"):
        ser_tmp = ser_resample.quantile(float(how[8:]))
    else:
        try:
            ser_tmp = getattr(ser_resample, how)()
        except AttributeError:
            ser_tmp = ser_resample.apply(how)
    fcov = (count / ref.resample(sampling).count()
            if minfrac is not None or how == "sum" else None)
    cond = fcov >= minfrac if minfrac is not None else count >= mincount
    ser_res = ser_tmp.reindex(cond.index)
    ser_res[~cond] = np.nan
    if how == "sum":
        ser_res = ser_res / fcov
    return ser_res


def resample_with_date(ser, ref, sampling, how, mincount=0, minfrac=None,
                       month_list=None):
    """Resample the given series and get the date of the max value.

    :param ser: a series containing the values
    :param ref: reference series
    :param sampling: aggregation type
    :param how: aggregation function
    :param mincount: minimum number of values that must be present to
                     accept a statistical result, most functions will
                     use ``minfrac`` instead
    :param minfrac: minimum fraction of data values that must be present
                    in order to accept the result of a statistical
                    function, if it is set to None ``mincount`` will be
                    used instead
    :param month_list: a list of months for ``seasonal`` statistics

    :return: A data frame with the resampled data
    """
    tmp = ser.fillna(-999.)
    grouped = tmp.groupby(pd.Grouper(freq=sampling))
    dfres = pd.DataFrame({"values": grouped.max(),
                          "date_of_maximum": grouped.idxmax()})
    if month_list is not None and 1 in month_list and 12 in month_list:
        dfres["date_of_maximum"] -= pd.Timedelta(182, "days")
    return dfres


def somo_processor(ser, ref, mtype, metadata, seasons, min_data_capture,
                   threshold, strict=False):
    """Calculate the sum of excess of daily maximum 8-hour means.

    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param metadata: metadata needed for some statistics
    :param seasons: season names
    :param min_data_capture: a fractional coverage value
    :param threshold: the cut-off value
    :param strict: a boolean denoting whether to use the strict rules
                   when to save a running mean

    :return: A list of dictionaries containing the processed data
    """
    label = f"somo{int(threshold)}{'_strict' if strict else ''}"
    tmp = dma8_processor(ser, ref, "daily", metadata, seasons,
                         min_data_capture, dname="EU", strict=strict)
    if mtype == "daily":
        res = tmp
        res[0]["name"] = label
    else:
        new_ser = tmp[0]["ser"] - threshold
        new_ser[new_ser < 0.] = 0.
        new_ref = pd.Series(1., tmp[0]["ref"].resample("D").asfreq().index)
        res = stat_processor_1(label, new_ser, new_ref, mtype, seasons,
                               how="sum", minfrac=min_data_capture)
    return res


def stat_processor_1(label, ser, ref, mtype, seasons, func=resample,
                     how="mean", mincount=0, minfrac=None, **kwargs):
    """Calculate statistics.

    :param label: key under which the results will be saved
    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param seasons: season names
    :param func: function to be called to actually calculate the
                 statistical quantity
    :param how: aggregation function
    :param mincount: minimum number of values that must be present to
                     accept a statistical result, most functions will
                     use ``minfrac`` instead
    :param minfrac: minimum fraction of data values that must be present
                    in order to accept the result of a statistical
                    function, if it is set to None ``mincount`` will be
                    used instead
    :param kwargs: any additional keyword arguments

    :return: A list of dictionaries containing the processed data
    """
    res = prepare_data(ser, ref, mtype, seasons, label)
    for r in res:
        r["ser"] = func(r["ser"], r["ref"], RSTAGS[mtype], how,
                        mincount=mincount, minfrac=minfrac,
                        month_list=r["month_list"], **kwargs)
    return res


def stat_processor_2(label, ser, ref, mtype, seasons, func=resample,
                     how="mean", mincount1=0, minfrac1=None, minfrac2=0.,
                     **kwargs):
    """Calculate statistics.

    First calculate ``daily`` statistics and then further process data
    for other than ``daily`` statistics.

    :param label: key under which the results will be saved
    :param ser: a series containing the values
    :param ref: reference series
    :param mtype: aggregation type
    :param seasons: season names
    :param func: function to be called to actually calculate the
                 statistical quantity
    :param how: aggregation function
    :param mincount1: minimum number of values that must be present to
                      accept a statistical result, most functions will
                      use ``minfrac`` instead
    :param minfrac1: minimum fraction of data values that must be
                     present in order to accept the result of a
                     statistical function, if it is set to None
                     ``mincount`` will be used instead
    :param minfrac2: minimum fraction of data values that must be
                     present in order to accept the result of a
                     statistical function, if it is set to None
                     ``mincount`` will be used instead
    :param kwargs: any additional keyword arguments

    :return: A list of dictionaries containing the processed data
    """
    ser_daily = func(ser, ref, "D", how, mincount=mincount1, minfrac=minfrac1,
                     **kwargs)
    ref_daily = ref.resample("D").count()
    res = prepare_data(ser_daily, ref_daily, mtype, seasons, label)
    if not mtype == "daily":
        for r in res:
            r["ser"] = resample(r["ser"], r["ref"], RSTAGS[mtype], how,
                                minfrac=minfrac2)
    return res


def threshold_processor(ser, ref, sampling, how, mincount=None, minfrac=0.75,
                        month_list=None, threshold=0.):
    """Evaluate counts above the given threshold.

    :param ser: a series containing the values
    :param ref: reference series
    :param sampling: aggregation type
    :param how: aggregation function
    :param mincount: minimum number of values that must be present to
                     accept a statistical result, most functions will
                     use ``minfrac`` instead
    :param minfrac: minimum fraction of data values that must be present
                    in order to accept the result of a statistical
                    function, if it is set to None ``mincount`` will be
                    used instead
    :param month_list: a list of months for ``seasonal`` statistics
    :param threshold: threshold value above which values shall be
                      accumulated

    :return: A series with the counts of values above the given
             threshold
    """
    fcov = (ser.dropna().resample(sampling).count()
            / ref.resample(sampling).count())
    cond = fcov >= minfrac
    ser_res = ser.where(ser > threshold).resample(sampling).count().reindex(
        cond.index
    )
    ser_res[~cond] = np.nan
    return ser_res


def w126stat(ser, ref, sampling, how, mincount=None, minfrac=0.75,
             month_list=None):
    """Evaluate the sum of values.

    Use a non-linear weighting function (W126 function).

    :param ser: a series containing the values
    :param ref: reference series
    :param sampling: aggregation type
    :param how: aggregation function
    :param mincount: minimum number of values that must be present to
                     accept a statistical result, most functions will
                     use ``minfrac`` instead
    :param minfrac: minimum fraction of data values that must be present
                    in order to accept the result of a statistical
                    function, if it is set to None ``mincount`` will be
                    used instead
    :param month_list: a list of months for ``seasonal`` statistics

    :return: A series with the sums of values
    """
    M = 4403.
    A = 126.
    ser_adj = ser/(1+M*np.exp(-A*ser/1000.))
    ser_adj_resample = ser_adj.resample(sampling)
    ser_adj_sum = ser_adj_resample.sum()
    count = ser_adj_resample.count()
    fcov = count / ref.resample(sampling).count()
    ser_res = ser_adj_sum / fcov
    ser_res[fcov < minfrac] = np.nan
    return ser_res.reindex(count.index)
