"""Ordinary least squares (OLS) linear regression calculation.

This module contains the following function:
ols - calculate the OLS linear regression
"""

import numpy as np
import scipy.stats
import statsmodels.formula.api as smf

from toarstats.trends.utils import moving_block_bootstrap


def ols(data):
    """Calculate the OLS linear regression.

    :param data: data containing a list of date time values and
                 associated parameter values on which to calculate the
                 trend

    :return: The trend with its uncertainty and p value
    """
    fit = smf.ols("value~datetime", data).fit(method="qr").params
    mbb = moving_block_bootstrap("OLS", data)
    fit_se = np.nanstd(mbb, axis=0)
    fit_pv = 2*scipy.stats.t.sf(x=abs(fit/fit_se), df=len(data)-2)
    return {"trend": fit, "uncertainty": fit_se, "p_value": fit_pv}
