"""This module contains helper functions for the trends subpackage.

This module contains the following function:
calculate_seasonal_cycle - calculate the seasonal cycle
calculate_anomalies - calculate the anomalies series
moving_block_bootstrap - perform the moving block bootstrap algorithm
"""

import numpy as np
import pandas as pd
import statsmodels.formula.api as smf


def calculate_seasonal_cycle(data):
    """Calculate the seasonal cycle.

    :param data: data containing a list of date time values and
                 associated parameter values

    :return: The seasonal cycle
    """
    return smf.ols(
        "value~np.sin(2*np.pi*month/12)+np.cos(2*np.pi*month/12)"
        "+np.sin(2*np.pi*month/6)+np.cos(2*np.pi*month/6)",
        pd.DataFrame({"value": data["value"], "month": data.index.month})
    ).fit(method="qr").predict(pd.DataFrame({"month": range(1, 13)}))


def calculate_anomalies(data):
    """Calculate the anomalies series.

    :param data: data containing a list of date time values and
                 associated parameter values

    :return: The anomalies series
    """
    seasonal_cycle = calculate_seasonal_cycle(data)
    idx = data.index
    return pd.DataFrame({
        "value": data["value"].values - seasonal_cycle[idx.month-1].values,
        "datetime": (idx.year-2000)*12 + idx.month
    }).sort_values("datetime")


def moving_block_bootstrap(method, data, quantile=None, num_samples=1000):
    """Perform the moving block bootstrap algorithm.

    :param method: either ``"OLS"`` or ``"quant"``
    :param data: data containing a list of date time values and
                 associated parameter values
    :param quantile: a single quantile, must be between 0 and 1 if
                     ``method="quant"``
    :param num_samples: number of sampled trends

    :return: A list of sampled trends
    """
    n = len(data)
    b = np.ceil(n**0.25).astype(int)
    nblocks = np.ceil(n/b).astype(int)
    blocks = np.array([list(range(i, i+b)) for i in range(n-b+1)])
    rng = np.random.default_rng()
    samples = []
    for _ in range(num_samples):
        bn = rng.choice(len(blocks), nblocks)
        samp_data = data.iloc[blocks[bn].flatten()]
        if method == "quant":
            mod = smf.quantreg("value~datetime", samp_data).fit(q=quantile)
        else:
            mod = smf.ols("value~datetime", samp_data).fit(method="qr")
        samples.append(mod.params)
    return samples
